#!/usr/bin/env python3
import random
import string
import sys

import django
from django.conf import settings
from django.test.utils import get_runner


SETTINGS = dict(
    SECRET_KEY=''.join([random.SystemRandom().choice(string.digits + string.ascii_letters + string.punctuation)
                        for i in range(50)]),
    DEBUG=True,
    INSTALLED_APPS=[
        'django.contrib.admin',
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'conditioner',
    ],
    TEMPLATES=[
        {
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'DIRS': [],
            'APP_DIRS': True,
            'OPTIONS': {
                'context_processors': [
                    'django.template.context_processors.debug',
                    'django.template.context_processors.request',
                    'django.contrib.auth.context_processors.auth',
                    'django.contrib.messages.context_processors.messages',
                ],
            },
        },
    ],
    DATABASES={
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': ':memory:',
        }
    },
)


def run_tests(*test_args):
    """
    Helper method for initializing dummy Django application and running 'django-conditioner' tests
    """
    if not test_args:
        test_args = ['conditioner']

    if not settings.configured:
        settings.configure(**SETTINGS)

    django.setup()
    TestRunner = get_runner(settings)
    test_runner = TestRunner()
    failures = test_runner.run_tests(test_args)

    sys.exit(bool(failures))


if __name__ == '__main__':
    run_tests(*sys.argv[1:])



"""
Conditioner module actions misc models
"""
import logging

from django.db import models

from conditioner.base import BaseAction


logger = logging.getLogger(__name__)


class LoggerAction(BaseAction):
    """
    Class representation of a logger action

    It takes an logger level and message, and logs it when linked condition is met.
    """
    DEBUG = 'DEBUG'
    INFO = 'INFO'
    WARNING = 'WARNING'
    ERROR = 'ERROR'
    CRITICAL = 'CRITICAL'
    LEVEL_CHOICES = (
        (DEBUG, 'Debug'),
        (INFO, 'Info'),
        (WARNING, 'Warning'),
        (ERROR, 'Error'),
        (CRITICAL, 'Critical'),
    )
    level = models.CharField(
        verbose_name='logging level',
        choices=LEVEL_CHOICES,
        max_length=64,
    )

    message = models.TextField(
        verbose_name='message',
    )

    class Meta(BaseAction.Meta):
        verbose_name = 'logger action'
        verbose_name_plural = 'logger actions'

    def run_action(self, *args, **kwargs):
        """
        Log saved message with saved logging level
        """
        logging_func = getattr(logger, self.level.lower())
        logging_func(self.message)

    def __str__(self):
        return 'Logger action ({0.level}: {0.message:.20})'.format(self)



from django.core.management.base import BaseCommand
from django.utils import timezone

from conditioner.base import BaseCronCondition


class Command(BaseCommand):
    help = "Check cron related conditions and run their actions if the condition is met"

    def handle(self, *args, **options):
        conditions = BaseCronCondition.objects.exclude(rule__isnull=True)

        for condition in conditions:
            # Make sure that there is an action to be run
            if not hasattr(condition.rule, 'action'):
                self.stdout.write(self.style.WARNING(
                    "Condition was met but rule {} doesn't have linked action.".format(condition.rule.pk))
                )
                continue

            # Generic conditions
            if condition.model_specific() is False:
                if condition.is_met():
                    condition.rule.action.run_action()
                    condition.last_executed = timezone.now()
                    condition.save()

                    self.stdout.write(self.style.SUCCESS(
                        "Condition for rule {} was met and the action "
                        "was successfully executed".format(condition.rule.pk))
                    )
            # Model specific conditions
            else:
                for instance in condition.rule.target_model.objects.all():
                    if condition.is_met(instance):
                        condition.rule.action.run_action(instance)
                        condition.last_executed = timezone.now()
                        condition.save()

                        self.stdout.write(self.style.SUCCESS(
                            "Model specific condition for rule {} was met and the action "
                            "was successfully executed".format(condition.rule.pk))
                        )

        self.stdout.write(self.style.SUCCESS("\nAll done!"))



"""
Test 'conditioner.models' file
"""
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.test import TestCase

from conditioner.models import Rule
from conditioner.tests.factories import RuleFactory, BaseActionFactory, BaseConditionFactory
from conditioner.utils import TimeStampedModelMixin


class RuleTestCase(TestCase):
    """
    Test `conditioner.Rule` model
    """
    def setUp(self):
        super().setUp()
        self.model = Rule
        self.instance = RuleFactory()

        # Link sample action and condition to rule
        BaseActionFactory(rule=self.instance)
        BaseConditionFactory(rule=self.instance)

    def test_model_inheritance(self):
        """Test model inheritance"""
        self.assertIsInstance(self.instance, models.Model)
        self.assertIsInstance(self.instance, TimeStampedModelMixin)

    def test_model_target_content_type_field(self):
        """Test model 'target_content_type' field"""
        field = self.model._meta.get_field('target_content_type')

        self.assertIsInstance(field, models.ForeignKey)
        self.assertEqual(field.rel.model, ContentType)
        self.assertEqual(field.rel.related_name, 'rules')
        self.assertEqual(field.verbose_name, 'target content type')
        self.assertTrue(field.null)
        self.assertTrue(field.blank)

    def test_model_meta_attributes(self):
        """Test model meta attributes"""
        meta = self.model._meta

        self.assertEqual(meta.verbose_name, 'rule')
        self.assertEqual(meta.verbose_name_plural, 'rules')

    def test_model_target_model_property(self):
        """Test model `target_model()` property"""
        self.assertEqual(
            self.instance.target_model,
            self.instance.target_content_type.model_class()
        )

    def test_model_str_method(self):
        """Test model `__str__` method"""
        self.assertIn(str(self.instance.action), str(self.instance))
        self.assertIn(str(self.instance.condition), str(self.instance))
        self.assertIn(str(self.instance.target_model), str(self.instance))



"""
Conditioner module Django Admin integration
"""
from django.contrib import admin
from django.contrib.admin.options import IS_POPUP_VAR
from django.contrib.contenttypes.models import ContentType

from polymorphic.admin import PolymorphicInlineSupportMixin, StackedPolymorphicInline

from conditioner.actions import SendTemplatedEmailAction, LoggerAction
from conditioner.actions.forms import SendTemplatedEmailActionModelForm
from conditioner.base import BaseAction, BaseCondition
from conditioner.conditions import DayOfMonthCondition, DayOfWeekCondition, ModelSignalCondition
from conditioner.models import Rule


class ActionInline(StackedPolymorphicInline):
    """
    An inline for `conditioner.base.BaseAction` polymorphic model

    By default allows default, already implemented actions but can be easily extended by adding
    `StackedPolymorphicInline.Child` classes with custom action models to 'child_inlines'
    """
    class SendTemplatedEmailActionInline(StackedPolymorphicInline.Child):
        model = SendTemplatedEmailAction
        form = SendTemplatedEmailActionModelForm

    class LoggerActionInline(StackedPolymorphicInline.Child):
        model = LoggerAction

    model = BaseAction
    child_inlines = [
        SendTemplatedEmailActionInline,
        LoggerActionInline,
    ]

    def get_formset_children(self, request, obj=None):
        """
        Extends default `get_formset_children()` behavior and filters available inline choices based on
        rule's `target_content_type` (if action `model_specific()` is specified).
        """
        formset_children = list()
        for child_inline in self.child_inline_instances:
            model_specific = child_inline.model.model_specific()

            # If rule has no target model and no model specific conditions should be visible
            if obj and not obj.target_model and model_specific:
                continue

            # If rule has a target model only those with matching model (or `True`, as it requires any model)
            # should be visible
            if (obj and obj.target_model and model_specific
                    and model_specific is not True
                    and model_specific is not obj.target_model):
                continue

            formset_children.append(child_inline.get_formset_child(request, obj=obj))

        return formset_children


class ConditionInline(StackedPolymorphicInline):
    """
    An inline for `conditioner.base.BaseCondition` polymorphic model

    By default allows default, already implemented conditions but can be easily extended by adding
    `StackedPolymorphicInline.Child` classes with custom conditions models to 'child_inlines'
    """
    class DayOfMonthConditionInline(StackedPolymorphicInline.Child):
        model = DayOfMonthCondition

    class DayOfWeekConditionInline(StackedPolymorphicInline.Child):
        model = DayOfWeekCondition

    class ModelSignalConditionInline(StackedPolymorphicInline.Child):
        model = ModelSignalCondition

    model = BaseCondition
    child_inlines = [
        DayOfMonthConditionInline,
        DayOfWeekConditionInline,
        ModelSignalConditionInline,
    ]

    def get_formset_children(self, request, obj=None):
        """
        Extends default `get_formset_children()` behavior and filters available inline choices based on
        rule's `target_content_type` (if condition `model_specific()` is specified).
        """
        formset_children = list()
        for child_inline in self.child_inline_instances:
            model_specific = child_inline.model.model_specific()

            # If rule has no target model and no model specific conditions should be visible
            if obj and not obj.target_model and model_specific:
                continue

            # If rule has a target model only those with matching model (or `True`, as it requires any model)
            # should be visible
            if (obj and obj.target_model and model_specific
                    and model_specific is not True
                    and model_specific is not obj.target_model):
                continue

            formset_children.append(child_inline.get_formset_child(request, obj=obj))

        return formset_children


@admin.register(Rule)
class RuleAdmin(PolymorphicInlineSupportMixin, admin.ModelAdmin):
    """
    Django Admin integration for `conditioner.Rule`model

    The 'action' and 'condition' relations are polymorphic, so we need `PolymorphicInlineSupportMixin` to make sure
    that inlines forms are properly handled
    """
    list_display = ('pk', 'target_content_type', 'action', 'condition', 'created', 'modified')

    def add_view(self, request, form_url='', extra_context=None):
        """
        Extends Django's default `add_view()` method and shows only 'target_content_type' field.Done that way to
        filter actions and conditions based on selected target content type (if they are model specific).
        """
        self.inlines = ()
        self.readonly_fields = ()
        return super().add_view(request, form_url, extra_context)

    def change_view(self, request, object_id, form_url='', extra_context=None):
        """
        Extends Django's default `add_view()` method, makes 'target_content_type' field readonly and adds inlines
        forms for rule action and condition. Done that way to filter actions and conditions based on selected
        target content type (if they are model specific).
        """
        self.inlines = (ActionInline, ConditionInline)
        self.readonly_fields = ('target_content_type',)
        return super().change_view(request, object_id, form_url, extra_context)

    def response_add(self, request, obj, post_url_continue=None):
        """
        Extends Django's default `response_add()` method and make 'Save' button act like 'Save and continue editing'.
        Done that way because create view doesn't show action and condition inline forms (change view does).
        Based on `django.contrib.auth.admin.UserAdmin.response_add()`
        """
        # The Save' button should behave like the 'Save and continue editing' button to allow adding actions and
        # conditions that are model related, except in two scenarios:
        #   - The user has pressed the 'Save and add another' button
        #   - We are adding in a popup
        if '_addanother' not in request.POST and IS_POPUP_VAR not in request.POST:
            request.POST['_continue'] = 1
        return super().response_add(request, obj, post_url_continue)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        """
        Extend Django's default `formfield_for_foreignkey()` method and filter available target content types.
        """
        if db_field.name == 'target_content_type':
            kwargs['queryset'] = ContentType.objects.exclude(
                # Let's avoid recursive rules for now
                app_label='conditioner'
            ).exclude(
                # Polymorphic base models shouldn't be used directly and only clutter the select list
                # To revisit in the future, as this is dependant on naming convention and thus isn't universal
                model__istartswith='base'
            ).exclude(
                # Exclude custom actions
                # As above, it's dependant on naming convention
                model__iendswith='action'
            ).exclude(
                # Exclude custom conditions
                # As above, it's dependant on naming convention
                model__iendswith='condition'
            )

        return super().formfield_for_foreignkey(db_field, request, **kwargs)



"""
Conditioner module common actions models
"""
from django.conf import settings
from django.core.mail import EmailMultiAlternatives
from django.db import models
from django.template import TemplateDoesNotExist
from django.template.loader import render_to_string

from conditioner.base import BaseAction
from conditioner.utils import get_available_templates


class SendTemplatedEmailAction(BaseAction):
    """
    Class representation of a send templated email action

    It takes an email subject and email template (selected from available templates) and sends it to passed
    email address when linked condition is met.
    """
    email = models.EmailField(
        verbose_name='email address',
    )

    subject = models.CharField(
        verbose_name='subject',
        max_length=256,
    )

    template = models.CharField(
        verbose_name='template',
        max_length=256,
    )

    class Meta(BaseAction.Meta):
        verbose_name = 'send templated email action'
        verbose_name_plural = 'send templated email actions'

    @staticmethod
    def available_templates():
        """
        Helper method for returning a list of available email templates. Template needs to be a `*.txt` file with
        a possible HTML counterpart.

        :return: list of available email templates paths
        :rtype: list of tuples
        """
        return get_available_templates('templates/conditioner/actions/emails/')

    def run_action(self, *args, **kwargs):
        """
        Send selected email template
        """
        email = EmailMultiAlternatives(
            subject=self.subject,
            body=render_to_string(self.template),
            from_email=settings.DEFAULT_FROM_EMAIL,
            to=[self.email],
        )

        # Attach HTML template with the same name if it exists
        try:
            email.attach_alternative(
                content=render_to_string(self.template.replace('.txt', '.html')),
                mimetype='text/html',
            )
        except TemplateDoesNotExist:
            pass

        return email.send()

    def __str__(self):
        return 'Send templated email action ({0.email}: {0.template})'.format(self)



"""
Test 'conditioner.conditions.dates' file
"""
from datetime import datetime

from django.core.validators import MaxValueValidator
from django.db import models
from django.test import TestCase

from freezegun import freeze_time

from conditioner.conditions.dates import DayOfMonthCondition, DayOfWeekCondition
from conditioner.base import BaseCronCondition
from conditioner.tests.conditions.factories import DayOfMonthConditionFactory, DayOfWeekConditionFactory


class DayOfMonthConditionTestCase(TestCase):
    """
    Test `conditioner.conditions.dates.DayOfMonthCondition` model
    """
    def setUp(self):
        super().setUp()
        self.model = DayOfMonthCondition
        self.instance = DayOfMonthConditionFactory()

    def test_model_inheritance(self):
        """Test model inheritance"""
        self.assertIsInstance(self.instance, BaseCronCondition)

    def test_model_day_field(self):
        """Test model 'day' field"""
        field = self.model._meta.get_field('day')

        self.assertIsInstance(field, models.PositiveSmallIntegerField)
        self.assertEqual(field.verbose_name, 'day of the month')
        self.assertEqual(field.help_text, "Action will occur every month on that day.")

    def test_model_day_field_validators(self):
        """Test model 'day' field validators"""
        field = self.instance._meta.get_field('day')
        validator = field.validators[0]

        self.assertIsInstance(validator, MaxValueValidator)
        self.assertEqual(validator.limit_value, 31)

    def test_model_meta_attributes(self):
        """Test model meta attributes"""
        meta = self.model._meta

        self.assertEqual(meta.verbose_name, 'day of month condition')
        self.assertEqual(meta.verbose_name_plural, 'day of month conditions')

    def test_model_is_met_method(self):
        """Test model `is_met()` method"""
        instance = DayOfMonthConditionFactory(day=2)

        # Wrong day
        with freeze_time('2016-01-01'):
            self.assertFalse(instance.is_met())

        # Correct day
        with freeze_time('2016-01-02'):
            self.assertTrue(instance.is_met())

        instance.last_executed = datetime(2016, 1, 2)
        instance.save()

        # Correct day, but the same date as 'last_executed'
        with freeze_time('2016-01-02'):
            self.assertFalse(instance.is_met())

        # Wrong day
        with freeze_time('2016-02-01'):
            self.assertFalse(instance.is_met())

        # Correct day
        with freeze_time('2016-02-02'):
            self.assertTrue(instance.is_met())

    def test_model_str_method(self):
        """Test model `__str__` method"""
        self.assertIn(str(self.instance.day), str(self.instance))


class DayOfWeekConditionTestCase(TestCase):
    """
    Test `conditioner.conditions.dates.DayOfWeekCondition` model
    """
    def setUp(self):
        super().setUp()
        self.model = DayOfWeekCondition
        self.instance = DayOfWeekConditionFactory()

    def test_model_inheritance(self):
        """Test model inheritance"""
        self.assertIsInstance(self.instance, BaseCronCondition)

    def test_model_weekday_field(self):
        """Test model 'weekday' field"""
        field = self.model._meta.get_field('weekday')

        self.assertIsInstance(field, models.PositiveSmallIntegerField)
        self.assertEqual(field.verbose_name, 'day of the week')
        self.assertEqual(field.choices, self.model.WEEKDAY_CHOICES)
        self.assertEqual(field.help_text, "Action will occur every week on that day.")

    def test_model_meta_attributes(self):
        """Test model meta attributes"""
        meta = self.model._meta

        self.assertEqual(meta.verbose_name, 'day of week condition')
        self.assertEqual(meta.verbose_name_plural, 'day of week conditions')

    def test_model_is_met_method(self):
        """Test model `is_met()` method"""
        instance = DayOfWeekConditionFactory(weekday=7)

        # Wrong weekday (January 1 2007 is Monday)
        with freeze_time('2007-01-01'):
            self.assertFalse(instance.is_met())

        # Correct weekday
        with freeze_time('2007-01-07'):
            self.assertTrue(instance.is_met())

        instance.last_executed = datetime(2007, 1, 7)
        instance.save()

        # Correct weekday, but the same date as 'last_executed'
        with freeze_time('2007-01-07'):
            self.assertFalse(instance.is_met())

        # Wrong weekday
        with freeze_time('2007-01-10'):
            self.assertFalse(instance.is_met())

        # Correct weekday
        with freeze_time('2007-01-14'):
            self.assertTrue(instance.is_met())

    def test_model_str_method(self):
        """Test model `__str__` method"""
        self.assertIn(self.instance.get_weekday_display(), str(self.instance))



"""
Conditioner module related utils
"""
import glob
import os

from django.db import models
from django.template.utils import get_app_template_dirs


class TimeStampedModelMixin(models.Model):
    """
    An abstract base class model that provides self managed 'created' and 'modified' fields.

    Based on:
        https://github.com/django-extensions/django-extensions/blob/master/django_extensions/db/models.py
    """
    created = models.DateTimeField(
        verbose_name='created',
        editable=False,
        auto_now_add=True,
    )

    modified = models.DateTimeField(
        verbose_name='modified',
        editable=False,
        auto_now=True,
    )

    class Meta:
        abstract = True
        get_latest_by = 'modified'
        ordering = ('-modified', '-created')


def get_available_templates(template_dir):
    """
    Helper method for returning a list of available email templates. Template needs to be a `*.txt` file and can have
    a HTML counterpart.

    :return: list of available email templates paths
    :rtype: list of tuples
    """
    templates_dir = get_app_template_dirs(template_dir)

    txt_templates = list()
    for template_dir in templates_dir:
        txt_templates += glob.glob(
            os.path.join(template_dir, '*.txt')
        )

    templates = list()
    for template in txt_templates:
        file_name, _ = os.path.splitext(os.path.basename(template))

        is_html_available = os.path.isfile(template.replace('.txt', '.html'))
        if is_html_available:
            label = '{} (txt + html)'.format(file_name)
        else:
            label = '{} (txt)'.format(file_name)

        templates.append(
            (template.split('templates/')[1], label)
        )

    return templates



"""
Conditioner
===========
Create simple 'if this then that' style rules in your Django application. Comes with a bunch of ready to use actions
and conditions, but is also easily extensible and allows model specific actions/conditions.
"""
default_app_config = 'conditioner.apps.ConditionerAppConfig'


__title__ = 'django-conditioner'
__version__ = '0.1.0'
__author__ = 'Omni Digital'
__license__ = 'MIT'
__url__ = 'https://github.com/omni-digital/django-conditioner'



"""
Test 'conditioner.utils' file
"""
from unittest import mock

from django.db import models
from django.test import TestCase

from conditioner.utils import TimeStampedModelMixin, get_available_templates


class TimeStampedModelMixinTestCase(TestCase):
    """
    Test `conditioner.utils.TimeStampedModelMixin` model
    """
    def setUp(self):
        super().setUp()
        self.model = TimeStampedModelMixin

    def test_model_inheritance(self):
        """Test model inheritance"""
        self.assertIsInstance(self.model(), models.Model)

    def test_model_created_field(self):
        """Test model 'created' field"""
        field = self.model._meta.get_field('created')

        self.assertIsInstance(field, models.DateTimeField)
        self.assertEqual(field.verbose_name, 'created')
        self.assertFalse(field.editable)
        self.assertTrue(field.auto_now_add)

    def test_model_modified_field(self):
        """Test model 'modified' field"""
        field = self.model._meta.get_field('modified')

        self.assertIsInstance(field, models.DateTimeField)
        self.assertEqual(field.verbose_name, 'modified')
        self.assertFalse(field.editable)
        self.assertTrue(field.auto_now)

    def test_model_meta_attributes(self):
        """Test model meta attributes"""
        meta = self.model._meta

        self.assertTrue(meta.abstract)
        self.assertEqual(meta.get_latest_by, 'modified')
        self.assertEqual(meta.ordering, ('-modified', '-created'))


class GetAvailableTemplatesTestCase(TestCase):
    """
    Test 'conditioner.utils.get_available_templates` function
    """

    @mock.patch('os.path.isfile')
    @mock.patch('glob.glob')
    @mock.patch('conditioner.utils.get_app_template_dirs')
    def test_get_available_templates_function(self, mocked_get_app_template_dirs, mocked_glob, mocked_is_file):
        """Test `conditioner.utils.get_available_templates` function"""
        mocked_get_app_template_dirs.return_value = ['/template/dir']
        mocked_glob.return_value = ['/path/templates/subdir/test.txt', '/path/templates/example.txt']
        mocked_is_file.side_effect = [True, False]

        templates = get_available_templates('templates/example/path/')
        expected_templates = [
            ('subdir/test.txt', 'test (txt + html)'),
            ('example.txt', 'example (txt)'),
        ]

        self.assertEqual(templates, expected_templates)
        self.assertEqual(mocked_is_file.call_count, 2)
        self.assertEqual(mocked_glob.call_count, 1)



"""
Conditioner module conditions signals related models
"""
import uuid

from django.db import models

from conditioner.base import BaseCondition


class ModelSignalCondition(BaseCondition):
    """
    Class representation of a model signal condition

    It mirrors available Django model signals and creates a given signal for linked rule content type, with
    linked action as signal receiver.
    """
    PRE_INIT = 'pre_init'
    POST_INIT = 'post_init'
    PRE_SAVE = 'pre_save'
    POST_SAVE = 'post_save'
    PRE_DELETE = 'pre_delete'
    POST_DELETE = 'post_delete'
    SIGNAL_CHOICES = (
        ('Save', (
            (PRE_SAVE, 'Before creation'),
            (POST_SAVE, 'After creation'),
        )),
        ('Delete', (
            (PRE_DELETE, 'Before deletion'),
            (POST_DELETE, 'After deletion'),
        )),
        ('Init', (
            (PRE_INIT, 'Before initialization'),
            (POST_INIT, 'After initialization'),
        )),
    )
    signal = models.CharField(
        verbose_name='signal',
        choices=SIGNAL_CHOICES,
        max_length=64,
    )

    dispatch_uid = models.UUIDField(
        verbose_name='dispatch ID',
        editable=False,
        default=uuid.uuid4,
    )

    class Meta(BaseCondition.Meta):
        verbose_name = 'model signal condition'
        verbose_name_plural = 'model signal conditions'

    @staticmethod
    def model_specific():
        """
        This action is model specific
        """
        return True

    def connect_signal(self):
        """
        Connect selected signal to rule's target model
        """
        if hasattr(self.rule, 'action'):
            signal = getattr(models.signals, self.signal)

            signal.connect(
                receiver=self.rule.action.run_action,
                sender=self.rule.target_model,
                dispatch_uid=self.dispatch_uid,
                weak=False,
            )

    def disconnects_signal(self):
        """
        Disconnect selected signal from rule's target model
        """
        if hasattr(self.rule, 'action'):
            signal = getattr(models.signals, self.signal)

            signal.disconnect(
                receiver=self.rule.action.run_action,
                sender=self.rule.target_model,
                dispatch_uid=self.dispatch_uid,
            )

    def __str__(self):
        return 'Model signal condition ({0.signal} for {0.rule.target_content_type})'.format(self)

    def save(self, *args, **kwargs):
        """
        Extends default `save()` behaviour and connects signal
        """
        if not self.pk:
            self.connect_signal()
        return super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        """
        Extends default `delete()` behaviour and disconnects signal
        """
        super_delete = super().delete(*args, **kwargs)
        self.disconnects_signal()  # After deletion to make sure that everything went OK
        return super_delete



"""
Test 'conditioner.apps' file
"""
from django.apps import AppConfig
from django.apps import apps as license_tracker_apps
from django.test import TestCase


class ConditionerAppConfigTestCase(TestCase):
    """
    Test 'conditioner' module `AppConfig` integration
    """
    def test_conditioner_app_config(self):
        """Test 'licenses' module `AppConfig` instance"""
        conditioner_app_config = license_tracker_apps.get_app_config('conditioner')

        self.assertIsInstance(conditioner_app_config, AppConfig)
        self.assertEqual(conditioner_app_config.name, 'conditioner')
        self.assertEqual(conditioner_app_config.verbose_name, 'Conditioner')



"""
Conditioner module conditions related factories
"""
import random

from conditioner.conditions import DayOfMonthCondition, DayOfWeekCondition, ModelSignalCondition
from conditioner.tests.factories import BaseConditionFactory, BaseCronConditionFactory


class DayOfMonthConditionFactory(BaseCronConditionFactory):
    """
    Factory for `conditioner.conditions.dates.DayOfMonthCondition` model
    """
    day = random.randint(1, 31)

    class Meta:
        model = DayOfMonthCondition


class DayOfWeekConditionFactory(BaseCronConditionFactory):
    """
    Factory for `conditioner.conditions.dates.DayOfWeekCondition` model
    """
    weekday = random.randint(1, 7)

    class Meta:
        model = DayOfWeekCondition


class ModelSignalConditionFactory(BaseConditionFactory):
    """
    Factory for `conditioner.conditions.signals.ModelSignalCondition` model
    """
    signal = random.choice([
        ModelSignalCondition.PRE_SAVE, ModelSignalCondition.POST_SAVE,
        ModelSignalCondition.PRE_DELETE, ModelSignalCondition.POST_DELETE,
        ModelSignalCondition.PRE_INIT, ModelSignalCondition.POST_INIT,
    ])

    class Meta:
        model = ModelSignalCondition



"""
Conditioner module related exceptions
"""


class BaseConditionerException(Exception):
    """
    Base class for all conditioner related exceptions
    """
    pass


class IncorrectModelException(BaseConditionerException):
    """
    Model specific action/condition got wrong object instance
    """
    pass



"""
Conditioner module base models
"""
from django.db import models

from polymorphic.models import PolymorphicModel


# Actions
class BaseAction(PolymorphicModel):
    """
    Class representation of a base action

    Should not be used directly and all actions must inherit from it.
    """
    rule = models.OneToOneField(
        'conditioner.Rule', related_name='action',
        verbose_name='rule',
    )

    @staticmethod
    def model_specific():
        """
        Returns model class (that inherits from `BaseAction`) if action is meant to be model specific and relies on
        model instances being passed to `run_action()`, `True` if it requires any model and `False` if the action is
        generic (default).
        """
        return False

    def run_action(self, *args, **kwargs):
        """
        Implements the actual action
        """
        raise NotImplementedError(
            "You have to implement 'run_action()' in all actions that inherit from 'BaseAction'"
        )

    def __str__(self):
        return 'Action'


# Conditions
class BaseCondition(PolymorphicModel):
    """
    Class representation of a base condition

    Should not be used directly and all conditions must inherit from it.
    """
    rule = models.OneToOneField(
        'conditioner.Rule', related_name='condition',
        verbose_name='rule',
    )

    @staticmethod
    def model_specific():
        """
        Returns model class (that inherits from `BaseCondition`) if action is meant to be model specific, `True` if it
        requires any model and `False` if the condition is generic (default).
        """
        return False

    def __str__(self):
        return 'Condition'


class BaseCronCondition(BaseCondition):
    """
    Class representation of a base cron condition

    Should not be used directly and all cron related conditions must inherit from it.
    """
    last_executed = models.DateTimeField(
        verbose_name='last executed',
        null=True,
        editable=False,
    )

    def is_met(self, *args, **kwargs):
        """
        Implements if the condition is met
        """
        raise NotImplementedError(
            "You have to implement 'is_met()' in all actions that inherit from 'BaseCronCondition'"
        )

    def __str__(self):
        return 'Cron condition'



"""
Conditioner module conditions

All available conditions should be imported here for ease of use.
"""
from conditioner.conditions.dates import DayOfMonthCondition, DayOfWeekCondition  # noqa
from conditioner.conditions.signals import ModelSignalCondition  # noqa



"""
Conditioner module actions

All available actions should be imported here for ease of use.
"""
from conditioner.actions.common import SendTemplatedEmailAction  # noqa
from conditioner.actions.misc import LoggerAction  # noqa



"""
Test 'conditioner.actions.misc' file
"""
from django.db import models
from django.test import TestCase

from conditioner.actions.misc import LoggerAction
from conditioner.base import BaseAction
from conditioner.tests.actions.factories import LoggerActionFactory


class LoggerActionTestCase(TestCase):
    """
    Test `conditioner.actions.misc.LoggerAction` model
    """
    def setUp(self):
        super().setUp()
        self.model = LoggerAction
        self.instance = LoggerActionFactory()

    def test_model_inheritance(self):
        """Test model inheritance"""
        self.assertIsInstance(self.instance, BaseAction)

    def test_model_level_field(self):
        """Test model 'level' field"""
        field = self.model._meta.get_field('level')

        self.assertIsInstance(field, models.CharField)
        self.assertEqual(field.verbose_name, 'logging level')
        self.assertEqual(field.choices, self.model.LEVEL_CHOICES)
        self.assertEqual(field.max_length, 64)

    def test_model_message_field(self):
        """Test model 'message' field"""
        field = self.model._meta.get_field('message')

        self.assertIsInstance(field, models.TextField)
        self.assertEqual(field.verbose_name, 'message')

    def test_model_meta_attributes(self):
        """Test model meta attributes"""
        meta = self.model._meta

        self.assertEqual(meta.verbose_name, 'logger action')
        self.assertEqual(meta.verbose_name_plural, 'logger actions')

    def test_model_run_action_method(self):
        """Test model `run_action()` method"""
        logger_name = 'conditioner.actions.misc'
        with self.assertLogs(logger_name, level='DEBUG') as log:
            self.instance.run_action()

        self.assertEqual(len(log.records), 1)

        log_item = log.records[0]
        self.assertEqual(self.instance.level, log_item.levelname)
        self.assertEqual(self.instance.message, log_item.message)

    def test_model_str_method(self):
        """Test model `__str__` method"""
        self.assertIn(self.instance.level, str(self.instance))
        self.assertIn(self.instance.message[:20], str(self.instance))



"""
Test 'conditioner.conditions.signals' file
"""
import uuid
from unittest import mock

from django.db import models
from django.test import TestCase

from conditioner.conditions.signals import ModelSignalCondition
from conditioner.base import BaseCondition
from conditioner.tests.conditions.factories import ModelSignalConditionFactory
from conditioner.tests.factories import RuleFactory, BaseActionFactory


class ModelSignalConditionTestCase(TestCase):
    """
    Test `conditioner.conditions.signals.ModelSignalCondition` model
    """
    def setUp(self):
        super().setUp()
        self.model = ModelSignalCondition
        self.instance = ModelSignalConditionFactory()

    def test_model_inheritance(self):
        """Test model inheritance"""
        self.assertIsInstance(self.instance, BaseCondition)

    def test_model_signal_field(self):
        """Test model 'signal' field"""
        field = self.model._meta.get_field('signal')

        self.assertIsInstance(field, models.CharField)
        self.assertEqual(field.verbose_name, 'signal')
        self.assertEqual(field.choices, self.model.SIGNAL_CHOICES)
        self.assertEqual(field.max_length, 64)

    def test_model_dispatch_uid_field(self):
        """Test model 'dispatch_uid' field"""
        field = self.model._meta.get_field('dispatch_uid')

        self.assertIsInstance(field, models.UUIDField)
        self.assertEqual(field.verbose_name, 'dispatch ID')
        self.assertFalse(field.editable)
        self.assertEqual(field.default, uuid.uuid4)

    def test_model_meta_attributes(self):
        """Test model meta attributes"""
        meta = self.model._meta

        self.assertEqual(meta.verbose_name, 'model signal condition')
        self.assertEqual(meta.verbose_name_plural, 'model signal conditions')

    @mock.patch('django.db.models.signals')
    def test_model_connect_signal_method(self, patched_signals):
        """Test model `connect_signal()` method"""
        mocked_signal = getattr(patched_signals, self.instance.signal)
        mocked_signal_connect = mock.Mock()
        mocked_signal.connect = mocked_signal_connect

        self.instance.rule = RuleFactory()
        self.instance.rule.action = BaseActionFactory()
        self.instance.connect_signal()

        mocked_signal_connect.assert_called_once_with(
            receiver=self.instance.rule.action.run_action,
            sender=self.instance.rule.target_model,
            dispatch_uid=self.instance.dispatch_uid,
            weak=False,
        )

    @mock.patch('django.db.models.signals')
    def test_model_disconnects_signal_method(self, patched_signals):
        """Test model `disconnects_signal()` method"""
        mocked_signal = getattr(patched_signals, self.instance.signal)
        mocked_signal_disconnect = mock.Mock()
        mocked_signal.disconnect = mocked_signal_disconnect

        self.instance.rule = RuleFactory()
        self.instance.rule.action = BaseActionFactory()
        self.instance.disconnects_signal()

        mocked_signal_disconnect.assert_called_once_with(
            receiver=self.instance.rule.action.run_action,
            sender=self.instance.rule.target_model,
            dispatch_uid=self.instance.dispatch_uid,
        )

    def test_model_str_method(self):
        """Test model `__str__` method"""
        self.assertIn(self.instance.signal, str(self.instance))
        self.assertIn(str(self.instance.rule.target_content_type), str(self.instance))

    @mock.patch('conditioner.conditions.signals.ModelSignalCondition.connect_signal')
    def test_model_save_method(self, mocked_connect_signal):
        """Test model `save()` method"""
        ModelSignalConditionFactory()

        self.assertEqual(mocked_connect_signal.call_count, 1)

    @mock.patch('conditioner.conditions.signals.ModelSignalCondition.disconnects_signal')
    def test_model_delete_method(self, mocked_disconnects_signal):
        """Test model `delete()` method"""
        ModelSignalConditionFactory().delete()

        self.assertEqual(mocked_disconnects_signal.call_count, 1)



"""
Test 'conditioner.actions.common' file
"""
from unittest import mock

from django.conf import settings
from django.db import models
from django.test import TestCase
from django.utils.crypto import get_random_string

from conditioner.actions.common import SendTemplatedEmailAction
from conditioner.base import BaseAction
from conditioner.tests.actions.factories import SendTemplatedEmailActionFactory


class LoggerActionTestCase(TestCase):
    """
    Test `conditioner.actions.common.SendTemplatedEmailAction` model
    """
    def setUp(self):
        super().setUp()
        self.model = SendTemplatedEmailAction
        self.instance = SendTemplatedEmailActionFactory()

    def test_model_inheritance(self):
        """Test model inheritance"""
        self.assertIsInstance(self.instance, BaseAction)

    def test_model_email_field(self):
        """Test model 'email' field"""
        field = self.model._meta.get_field('email')

        self.assertIsInstance(field, models.EmailField)
        self.assertEqual(field.verbose_name, 'email address')

    def test_model_subject_field(self):
        """Test model 'subject' field"""
        field = self.model._meta.get_field('subject')

        self.assertIsInstance(field, models.CharField)
        self.assertEqual(field.verbose_name, 'subject')
        self.assertEqual(field.max_length, 256)

    def test_model_template_field(self):
        """Test model 'template' field"""
        field = self.model._meta.get_field('template')

        self.assertIsInstance(field, models.CharField)
        self.assertEqual(field.verbose_name, 'template')
        self.assertEqual(field.max_length, 256)

    def test_model_meta_attributes(self):
        """Test model meta attributes"""
        meta = self.model._meta

        self.assertEqual(meta.verbose_name, 'send templated email action')
        self.assertEqual(meta.verbose_name_plural, 'send templated email actions')

    @mock.patch('conditioner.actions.common.get_available_templates')
    def test_model_available_templates_method(self, mocked_get_available_templates):
        """Test model `get_available_templates()` method"""
        self.instance.available_templates()

        mocked_get_available_templates.assert_called_once_with(
            'templates/conditioner/actions/emails/'
        )

    @mock.patch('conditioner.actions.common.EmailMultiAlternatives')
    @mock.patch('conditioner.actions.common.render_to_string')
    def test_model_run_action_method(self, mocked_render_to_string, mocked_email):
        """Test model `run_action()` method"""
        rendered_template = get_random_string()
        mocked_render_to_string.return_value = rendered_template

        self.instance.run_action()

        mocked_email.assert_called_once_with(
            subject=self.instance.subject,
            body=rendered_template,
            from_email=settings.DEFAULT_FROM_EMAIL,
            to=[self.instance.email],
        )

        self.assertEqual(mocked_render_to_string.call_count, 2)
        mocked_render_to_string.assert_any_call(self.instance.template)
        mocked_render_to_string.assert_any_call(self.instance.template.replace('.txt', '.html'))

    def test_model_str_method(self):
        """Test model `__str__` method"""
        self.assertIn(self.instance.email, str(self.instance))
        self.assertIn(self.instance.template, str(self.instance))



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-16 15:29
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='BaseAction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BaseCondition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Rule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
                ('target_content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='rules', to='contenttypes.ContentType', verbose_name='target content type')),
            ],
            options={
                'ordering': ('-modified', '-created'),
                'abstract': False,
                'get_latest_by': 'modified',
                'verbose_name': 'rule',
                'verbose_name_plural': 'rules',
            },
        ),
        migrations.CreateModel(
            name='BaseCronCondition',
            fields=[
                ('basecondition_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='conditioner.BaseCondition')),
                ('last_executed', models.DateTimeField(editable=False, null=True, verbose_name='last executed')),
            ],
            options={
                'abstract': False,
            },
            bases=('conditioner.basecondition',),
        ),
        migrations.CreateModel(
            name='LoggerAction',
            fields=[
                ('baseaction_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='conditioner.BaseAction')),
                ('level', models.CharField(choices=[('DEBUG', 'Debug'), ('INFO', 'Info'), ('WARNING', 'Warning'), ('ERROR', 'Error'), ('CRITICAL', 'Critical')], max_length=64, verbose_name='logging level')),
                ('message', models.TextField(verbose_name='message')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'logger action',
                'verbose_name_plural': 'logger actions',
            },
            bases=('conditioner.baseaction',),
        ),
        migrations.CreateModel(
            name='ModelSignalCondition',
            fields=[
                ('basecondition_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='conditioner.BaseCondition')),
                ('signal', models.CharField(choices=[('Save', (('pre_save', 'Before creation'), ('post_save', 'After creation'))), ('Delete', (('pre_delete', 'Before deletion'), ('post_delete', 'After deletion'))), ('Init', (('pre_init', 'Before initialization'), ('post_init', 'After initialization')))], max_length=64, verbose_name='signal')),
                ('dispatch_uid', models.UUIDField(default=uuid.uuid4, editable=False, verbose_name='dispatch ID')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'model signal condition',
                'verbose_name_plural': 'model signal conditions',
            },
            bases=('conditioner.basecondition',),
        ),
        migrations.CreateModel(
            name='SendTemplatedEmailAction',
            fields=[
                ('baseaction_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='conditioner.BaseAction')),
                ('email', models.EmailField(max_length=254, verbose_name='email address')),
                ('subject', models.CharField(max_length=256, verbose_name='subject')),
                ('template', models.CharField(max_length=256, verbose_name='template')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'send templated email action',
                'verbose_name_plural': 'send templated email actions',
            },
            bases=('conditioner.baseaction',),
        ),
        migrations.AddField(
            model_name='basecondition',
            name='polymorphic_ctype',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_conditioner.basecondition_set+', to='contenttypes.ContentType'),
        ),
        migrations.AddField(
            model_name='basecondition',
            name='rule',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='condition', to='conditioner.Rule', verbose_name='rule'),
        ),
        migrations.AddField(
            model_name='baseaction',
            name='polymorphic_ctype',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_conditioner.baseaction_set+', to='contenttypes.ContentType'),
        ),
        migrations.AddField(
            model_name='baseaction',
            name='rule',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='action', to='conditioner.Rule', verbose_name='rule'),
        ),
        migrations.CreateModel(
            name='DayOfMonthCondition',
            fields=[
                ('basecroncondition_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='conditioner.BaseCronCondition')),
                ('day', models.PositiveSmallIntegerField(help_text='Action will occur every month on that day.', validators=[django.core.validators.MaxValueValidator(31)], verbose_name='day of the month')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'day of month condition',
                'verbose_name_plural': 'day of month conditions',
            },
            bases=('conditioner.basecroncondition',),
        ),
        migrations.CreateModel(
            name='DayOfWeekCondition',
            fields=[
                ('basecroncondition_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='conditioner.BaseCronCondition')),
                ('weekday', models.PositiveSmallIntegerField(choices=[(1, 'Monday'), (2, 'Tuesday'), (3, 'Wednesday'), (4, 'Thursday'), (5, 'Friday'), (6, 'Sunday'), (7, 'Saturday')], help_text='Action will occur every week on that day.', verbose_name='day of the week')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'day of week condition',
                'verbose_name_plural': 'day of week conditions',
            },
            bases=('conditioner.basecroncondition',),
        ),
    ]



"""
Conditioner module actions related forms
"""
from django import forms

from conditioner.actions.common import SendTemplatedEmailAction


class SendTemplatedEmailActionModelForm(forms.ModelForm):
    """
    Model form for `conditioner.actions.common.SendTemplatedEmailAction`
    """
    def __init__(self, *args, **kwargs):
        """
        Extends Django's default initialization and adds choices to 'template' field
        """
        super().__init__(*args, **kwargs)
        self.fields['template'].widget = forms.Select(choices=SendTemplatedEmailAction.available_templates())

    class Meta:
        model = SendTemplatedEmailAction
        fields = '__all__'



"""
Conditioner module related factories
"""
from django.contrib.contenttypes.models import ContentType

import factory
from faker import Factory as FakerFactory

from conditioner.base import BaseAction, BaseCondition, BaseCronCondition


faker = FakerFactory.create()


class RuleFactory(factory.DjangoModelFactory):
    """
    Factory for `conditioner.Rule` model
    """
    target_content_type = factory.Iterator(ContentType.objects.all())

    class Meta:
        model = 'conditioner.Rule'


class BaseActionFactory(factory.DjangoModelFactory):
    """
    Factory for `conditioner.base.BaseAction` model
    """
    rule = factory.SubFactory(RuleFactory)

    class Meta:
        model = BaseAction


class BaseConditionFactory(factory.DjangoModelFactory):
    """
    Factory for `conditioner.base.BaseCondition` model
    """
    rule = factory.SubFactory(RuleFactory)

    class Meta:
        model = BaseCondition


class BaseCronConditionFactory(BaseConditionFactory):
    """
    Factory for `conditioner.base.BaseCronCondition` model
    """

    class Meta:
        model = BaseCronCondition



"""
Test 'conditioner.base' file
"""
from django.db import models
from django.test import TestCase

from polymorphic.models import PolymorphicModel

from conditioner.base import BaseAction, BaseCondition, BaseCronCondition
from conditioner.models import Rule
from conditioner.tests.factories import BaseActionFactory, BaseConditionFactory, BaseCronConditionFactory


class BaseActionTestCase(TestCase):
    """
    Test `conditioner.base.BaseAction` model
    """
    def setUp(self):
        super().setUp()
        self.model = BaseAction
        self.instance = BaseActionFactory()

    def test_model_inheritance(self):
        """Test model inheritance"""
        self.assertIsInstance(self.instance, models.Model)
        self.assertIsInstance(self.instance, PolymorphicModel)

    def test_model_rule_field(self):
        """Test model 'rule' field"""
        field = self.model._meta.get_field('rule')

        self.assertIsInstance(field, models.OneToOneField)
        self.assertEqual(field.rel.model, Rule)
        self.assertEqual(field.rel.related_name, 'action')
        self.assertEqual(field.verbose_name, 'rule')

    def test_model_run_action_method(self):
        """Test model `run_action()` method"""
        self.assertRaises(NotImplementedError, self.instance.run_action)

    def test_model_str_method(self):
        """Test model `__str__` method"""
        self.assertEqual(str(self.instance), 'Action')


class BaseConditionTestCase(TestCase):
    """
    Test `conditioner.base.BaseCondition` model
    """
    def setUp(self):
        super().setUp()
        self.model = BaseCondition
        self.instance = BaseConditionFactory()

    def test_model_inheritance(self):
        """Test model inheritance"""
        self.assertIsInstance(self.instance, models.Model)
        self.assertIsInstance(self.instance, PolymorphicModel)

    def test_model_rule_field(self):
        """Test model 'rule' field"""
        field = self.model._meta.get_field('rule')

        self.assertIsInstance(field, models.OneToOneField)
        self.assertEqual(field.rel.model, Rule)
        self.assertEqual(field.rel.related_name, 'condition')
        self.assertEqual(field.verbose_name, 'rule')

    def test_model_str_method(self):
        """Test model `__str__` method"""
        self.assertEqual(str(self.instance), 'Condition')


class BaseCronConditionTestCase(TestCase):
    """
    Test `conditioner.base.BaseCronCondition` model
    """
    def setUp(self):
        super().setUp()
        self.model = BaseCronCondition
        self.instance = BaseCronConditionFactory()

    def test_model_inheritance(self):
        """Test model inheritance"""
        self.assertIsInstance(self.instance, BaseCondition)

    def test_model_last_executed_field(self):
        """Test model 'last_executed' field"""
        field = self.model._meta.get_field('last_executed')

        self.assertIsInstance(field, models.DateTimeField)
        self.assertEqual(field.verbose_name, 'last executed')
        self.assertTrue(field.null)
        self.assertFalse(field.editable)

    def test_model_str_method(self):
        """Test model `__str__` method"""
        self.assertEqual(str(self.instance), 'Cron condition')



import io
import os
import re

from setuptools import setup, find_packages


# Convert description from markdown to reStructuredText
try:
    import pypandoc
    description = pypandoc.convert('README.md', 'rst', 'markdown')
except (OSError, ImportError):
    description = ''


# Get package version number
# Source: https://packaging.python.org/single_source_version/
def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='django-conditioner',
    url='https://github.com/omni-digital/django-conditioner',
    download_url='https://github.com/omni-digital/django-conditioner/releases/latest',
    bugtrack_url='https://github.com/omni-digital/django-conditioner/issues',
    version=find_version('conditioner', '__init__.py'),
    license='MIT License',
    author='Omni Digital',
    author_email='dev@omni-digital.co.uk',
    maintainer='Omni Digital',
    maintainer_email='dev@omni-digital.co.uk',
    description="Create simple 'if this then that' style rules in your Django application. Comes with a bunch of "
                "ready to use actions and conditions, but is also easily extensible and allows model specific "
                "actions/conditions.",
    long_description=description,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'django-polymorphic>=1.1',
    ],
    test_suite='runtests.run_tests',
    tests_require=[
        'factory_boy>=2.8.1',
        'freezegun>=0.3.8',
    ],
    keywords='django conditions ifttt',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



"""
Date related conditions models
"""
from django.core.validators import MaxValueValidator
from django.db import models
from django.utils import timezone

from conditioner.base import BaseCronCondition


class DayOfMonthCondition(BaseCronCondition):
    """
    Class representation of a day of month condition

    It allows for date based conditions like 'every first day of the month'.
    """
    day = models.PositiveSmallIntegerField(
        verbose_name='day of the month',
        validators=[MaxValueValidator(31)],
        help_text="Action will occur every month on that day.",
    )

    class Meta(BaseCronCondition.Meta):
        verbose_name = 'day of month condition'
        verbose_name_plural = 'day of month conditions'

    def is_met(self, *args, **kwargs):
        """
        Condition is met when current date day is equal to `value` and `last_executed` isn't today
        """
        today = timezone.now().date()
        if self.day == today.day and (not self.last_executed or self.last_executed.date() != today):
            return True

        return False

    def __str__(self):
        return 'Day of month condition (day: {0.day})'.format(self)


class DayOfWeekCondition(BaseCronCondition):
    """
    Class representation of a day of week condition

    It allows for date based conditions like 'every Monday'.
    """
    # Based on ISO weekdays values
    WEEKDAY_CHOICES = (
        (1, 'Monday'),
        (2, 'Tuesday'),
        (3, 'Wednesday'),
        (4, 'Thursday'),
        (5, 'Friday'),
        (6, 'Sunday'),
        (7, 'Saturday'),
    )
    weekday = models.PositiveSmallIntegerField(
        verbose_name='day of the week',
        choices=WEEKDAY_CHOICES,
        help_text="Action will occur every week on that day.",
    )

    class Meta(BaseCronCondition.Meta):
        verbose_name = 'day of week condition'
        verbose_name_plural = 'day of week conditions'

    def is_met(self, *args, **kwargs):
        """
        Condition is met when current date week day is equal to `day` and `last_executed` isn't today
        """
        today = timezone.now().date()
        if self.weekday == today.isoweekday() and (not self.last_executed or self.last_executed.date() != today):
            return True

        return False

    def __str__(self):
        return 'Day of week condition (day: {0})'.format(self.get_weekday_display())



"""
Test 'conditioner.actions.forms' file
"""
from django import forms
from django.test import TestCase

from conditioner.actions import SendTemplatedEmailAction
from conditioner.actions.forms import SendTemplatedEmailActionModelForm


class SendTemplatedEmailActionModelFormTestCase(TestCase):
    """
    Test `conditioner.actions.forms.SendTemplatedEmailActionModelForm` form
    """
    def setUp(self):
        super().setUp()
        self.form = SendTemplatedEmailActionModelForm
        self.instance = SendTemplatedEmailActionModelForm()

    def test_form_inheritance(self):
        """Test form inheritance"""
        self.assertIsInstance(self.instance, forms.ModelForm)

    def test_form_template_field_choices(self):
        """Test form 'template' field choices"""
        self.assertIsInstance(self.instance.fields['template'].widget, forms.Select)
        self.assertEqual(
            self.instance.fields['template'].widget.choices,
            SendTemplatedEmailAction.available_templates()
        )

    def test_form_meta_attributes(self):
        """Test form meta attributes"""
        meta = self.form._meta

        self.assertEqual(meta.model, SendTemplatedEmailAction)
        self.assertIsNone(meta.fields)  # '__all__' is translated to None in the background



"""
Conditioner module AppConfig integration
"""
from django import db
from django.apps import AppConfig


class ConditionerAppConfig(AppConfig):
    """
    Django AppConfig integration for `conditioner` module
    """
    name = 'conditioner'
    verbose_name = 'Conditioner'

    def ready(self):
        # Make sure that all models are imported
        from conditioner import actions  # noqa
        from conditioner import conditions  # noqa

        # Signals aren't persistent so we need to register them on startup
        try:
            for signal_condition in conditions.ModelSignalCondition.objects.all():
                signal_condition.connect_signal()
        except db.DatabaseError:  # Django migrations weren't run yet
            pass



"""
Conditioner module actions related factories
"""
import random

import factory
from faker import Factory as FakerFactory

from conditioner.actions import LoggerAction, SendTemplatedEmailAction
from conditioner.tests.factories import BaseActionFactory


faker = FakerFactory.create()


class LoggerActionFactory(BaseActionFactory):
    """
    Factory for `conditioner.actions.misc.LoggerAction` model
    """
    level = random.choice(LoggerAction.LEVEL_CHOICES)[0]
    message = factory.LazyAttribute(lambda n: faker.paragraph())

    class Meta:
        model = LoggerAction


class SendTemplatedEmailActionFactory(BaseActionFactory):
    """
    Factory for `conditioner.actions.common.SendTemplatedEmailAction` model
    """
    email = factory.LazyAttribute(lambda n: faker.email())
    subject = factory.LazyAttribute(lambda n: faker.sentence())
    template = factory.LazyAttribute(lambda n: faker.uri_path() + '.txt')

    class Meta:
        model = SendTemplatedEmailAction



"""
Conditioner module models
"""
from django.db import models
from django.contrib.contenttypes.models import ContentType

from conditioner.utils import TimeStampedModelMixin


class Rule(TimeStampedModelMixin, models.Model):
    """
    Class representation of generic rule with target, condition and action
    """
    target_content_type = models.ForeignKey(
        ContentType, related_name='rules',
        verbose_name='target content type',
        null=True,
        blank=True,
    )

    # There are reverse 'BaseAction' and 'BaseCondition' foreign keys to allow Django Admin inline form
    # handling (with polymorphism), and thus aren't enforced but *should* be considered required

    class Meta(TimeStampedModelMixin.Meta):
        verbose_name = 'rule'
        verbose_name_plural = 'rules'

    @property
    def target_model(self):
        """Helper property for returning target model class"""
        return self.target_content_type.model_class() if self.target_content_type else None

    def __str__(self):
        if hasattr(self, 'action') and hasattr(self, 'condition'):
            return "Do '{0.action}' to '{0.target_model}' when '{0.condition}'".format(self)

        return 'Rule for {0.target_content_type}'.format(self)
